'''

will use https://github.com/stanstrup/commonMZ for now, which come mostly from Keller et al, 2008.

Could consider the tables from CAMERA later.

Two sets of tables for pos and neg ESI.

'''
















